//GMRES and CG simulations
//with compression

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <mpi.h>
#include <unistd.h>
#include <sys/time.h>
#include "cublas.h"
#include <cutil.h>
#include "tests.h"
#include <cmath>
#include <cutil_inline.h>
#include "iohb.h"
#include "mmio.h"
#include <ctype.h>
#include <vector>
#include <algorithm>
#include "utils.h"
#include "texture.h"

//Scalar variables
int restart = 16;
double precision = 1e-12;
int maxiter = 500;
int rang = 0;
int symmetric = 0;
int nbprocs;
int rank;
int method = 0;
int sizeloc;
int offset;
int sizeglob;
int CutRow, LeftPortion, RightPortion, LeftColumn, RightColumn, LeftEcart, RightEcart;
int *neighbors, *neighborr;
int XSize;
int iteration;
int nombre=1;

//sparse matrix 
typedef struct{
	int NumRows;  
	int NumColumns;
	int NumNonZeros;
	int *Pointers;  
	int *Columns;   
	double *NonZeros;
	int Bandwidth;
	int *TabSizes;
	int *Taboffsets;
	int *sendcount;
	int *sdispls;
	int *recvcount;
	int *rdispls;
	int ranksend;
	int rankrecv;
	int *RankSend;
	int *RankRecv;
	int *RecvColumns;
	int *SendColumns;
	int send;
	int recv;
	int tag;
	MPI_Request *SendReq;
	MPI_Request *RecvReq;
	double *CompressedX;
	int nmax;
}SparseMatrix;

SparseMatrix MatrixA, MatrixB, Matrix, matrix;

texture<int2,1> TextureX;

//---------------------------	FOR MATRIX GENERATION	--------------------------------//
//MAX NUMBER OF NONZEROS IN A MATRIX ROW
void max_nonzeros(SparseMatrix *matrix)
{
	int i, nmax;
	
	matrix->nmax = 0;

	for(i=0; i<matrix->NumRows; i++){
		nmax = matrix->Pointers[i+1] - matrix->Pointers[i];
		if(nmax > matrix->nmax)
			matrix->nmax = nmax;
	}
}

//CONVERSION FROM CSR TO CSC OR VICE VERSA
void Conversion_CSR_CSC(SparseMatrix *InMatrix, SparseMatrix *OutMatrix)
{
	int i, j, col, index;
	int *vect;

	OutMatrix->NumRows     = InMatrix->NumColumns;
	OutMatrix->NumColumns  = InMatrix->NumRows;
	OutMatrix->NumNonZeros = InMatrix->NumNonZeros;

	OutMatrix->Pointers = (int*)malloc(sizeof(int)*(OutMatrix->NumRows+1));
	OutMatrix->Columns  = (int*)malloc(sizeof(int)*OutMatrix->NumNonZeros);
	OutMatrix->NonZeros = (double*)malloc(sizeof(double)*OutMatrix->NumNonZeros);

	vect = (int*)malloc(sizeof(int)*(OutMatrix->NumRows+1));
	for(i=0; i<OutMatrix->NumRows+1; i++) vect[i] = 0;
	for(i=0; i<InMatrix->NumNonZeros; i++) vect[InMatrix->Columns[i]] ++;

	OutMatrix->Pointers[0] = 0;
	for(i=0; i<OutMatrix->NumRows; i++){
		OutMatrix->Pointers[i+1] = OutMatrix->Pointers[i] + vect[i];
		vect[i] = OutMatrix->Pointers[i];
	}
	vect[OutMatrix->NumRows] = OutMatrix->Pointers[OutMatrix->NumRows];

	for(i=0; i<InMatrix->NumRows; i++){
		for(j=InMatrix->Pointers[i]; j<InMatrix->Pointers[i+1]; j++){
			col = InMatrix->Columns[j];
			index = vect[col];
			OutMatrix->NonZeros[index] = InMatrix->NonZeros[j];
			OutMatrix->Columns[index] = i;
			vect[col] ++;		
		}
	}

	free(vect);
	free(InMatrix->Columns);
	free(InMatrix->Pointers);
	free(InMatrix->NonZeros);
}

//LOADED MATRIX
void Loaded_Matrix(SparseMatrix *Matrix, SparseMatrix *Upper, SparseMatrix *Lower, int symmetric)
{
	int i, j, col, index, nb, count;
	int *vect;

//CSC TO CSR
	Lower->NumRows = Upper->NumColumns;
	Lower->NumColumns = Upper->NumRows;
	Lower->NumNonZeros = Upper->NumNonZeros;

	Lower->Pointers = (int*)malloc(sizeof(int)*(Lower->NumRows+1));
	Lower->Columns = (int*)malloc(sizeof(int)*Lower->NumNonZeros);
	Lower->NonZeros = (double*)malloc(sizeof(double)*Lower->NumNonZeros);

	vect = (int*)malloc(sizeof(int)*(Lower->NumRows+1));
	for(i=0; i<Lower->NumRows+1; i++) vect[i] = 0;
	for(i=0; i<Upper->NumNonZeros; i++) vect[Upper->Columns[i]] ++;

	Lower->Pointers[0] = 0;
	for(i=0; i<Lower->NumRows; i++){
		Lower->Pointers[i+1] = Lower->Pointers[i] + vect[i];
		vect[i] = Lower->Pointers[i];
	}
	vect[Lower->NumRows] = Lower->Pointers[Lower->NumRows];

	for(i=0; i<Upper->NumRows; i++){
		for(j=Upper->Pointers[i]; j<Upper->Pointers[i+1]; j++){
			col = Upper->Columns[j];
			index = vect[col];
			Lower->NonZeros[index] = Upper->NonZeros[j];
			Lower->Columns[index] = i;
			vect[col] ++;		
		}
	}
	free(vect);

//SYMMETRIC
	if(symmetric){
		Matrix->NumRows = Lower->NumRows;
		Matrix->NumColumns = Lower->NumColumns;
		Matrix->NumNonZeros = Lower->NumNonZeros + Upper->NumNonZeros - Lower->NumRows;

		Matrix->Pointers = (int*)malloc(sizeof(int)*(Matrix->NumRows+1));
		Matrix->Columns = (int*)malloc(sizeof(int)*Matrix->NumNonZeros);
		Matrix->NonZeros = (double*)malloc(sizeof(double)*Matrix->NumNonZeros);

		Matrix->Pointers[0] = 0;
		nb = 0;

		for(i=0; i<Matrix->NumRows; i++){
			count = 0;
			for(j=Lower->Pointers[i]; j<Lower->Pointers[i+1]; j++){
				Matrix->Columns[nb] = Lower->Columns[j];
				Matrix->NonZeros[nb] = Lower->NonZeros[j];
				nb ++;
				count ++;
			}

			nb --;
			Matrix->NonZeros[nb] += Upper->NonZeros[Upper->Pointers[i]];
			nb ++;

			for(j=Upper->Pointers[i]+1; j<Upper->Pointers[i+1]; j++){
				Matrix->Columns[nb] = Upper->Columns[j];
				Matrix->NonZeros[nb] = Upper->NonZeros[j];
				nb ++;
				count ++;
			}

			Matrix->Pointers[i+1] = Matrix->Pointers[i] + count;
		}
	}

//UNSYMMETRIC
	else{
		Matrix->NumRows = Lower->NumRows;
		Matrix->NumColumns = Lower->NumColumns;
		Matrix->NumNonZeros = Lower->NumNonZeros;

		Matrix->Pointers = (int*)malloc(sizeof(int)*(Matrix->NumRows+1));
		Matrix->Columns = (int*)malloc(sizeof(int)*Matrix->NumNonZeros);
		Matrix->NonZeros = (double*)malloc(sizeof(double)*Matrix->NumNonZeros);

		for(i=0; i<Matrix->NumRows+1; i++) Matrix->Pointers[i] = Lower->Pointers[i];
		for(i=0; i<Matrix->NumNonZeros; i++){
			Matrix->Columns[i] = Lower->Columns[i];
			Matrix->NonZeros[i] = Lower->NonZeros[i];
		}

	}

//CLEAN
	free(Lower->Pointers);
	free(Lower->Columns);
	free(Lower->NonZeros);
	free(Upper->Pointers);
	free(Upper->Columns);
	free(Upper->NonZeros);
}

//BANDWIDTH OF THE MATRIX
void Bandwidth(SparseMatrix *Matrix)
{
	int i, width;

	Matrix->Bandwidth = 0;
	
	for(i=0; i<Matrix->NumRows; i++){
		width = Matrix->Columns[Matrix->Pointers[i+1]-1] - Matrix->Columns[Matrix->Pointers[i]] + 1; 
		if(width > Matrix->Bandwidth)
			Matrix->Bandwidth = width;
	}
}

//LARGER ROW
void LargerRow(SparseMatrix *Matrix)
{
	int i, width, middle, bandwidth;

	middle = Matrix->NumRows / 2;
	bandwidth = 0;
	
	for(i=0; i<Matrix->NumRows; i++){
		width = Matrix->Columns[Matrix->Pointers[i+1]-1] - Matrix->Columns[Matrix->Pointers[i]] + 1; 
		if(width >= bandwidth){
			if(abs(middle-i) < abs(middle-CutRow))
				CutRow = i;
			bandwidth = width;
		}
	}
	
}

//Compute the sizes of the Left_Part and the Right_Part sub-copies of the loaded sparse matrix
void Size_Sub_Copies(SparseMatrix *Matrix)
{
	int i, j;
	
	LeftPortion = 0;
	RightPortion = 0;
	LeftColumn = Matrix->NumRows - 1;
	RightColumn = 0;
	
	//Left
	for(i=CutRow; i<Matrix->NumRows; i++){
		for(j=Matrix->Pointers[i]; j<Matrix->Pointers[i+1]; j++)
			if(Matrix->Columns[j] <= CutRow)
				LeftPortion ++;
	
		if(LeftColumn > Matrix->Columns[Matrix->Pointers[i]])
			LeftColumn = Matrix->Columns[Matrix->Pointers[i]];
	}
	
	//Right
	for(i=CutRow; i>=0; i--){
		for(j=Matrix->Pointers[i]; j<Matrix->Pointers[i+1]; j++)
			if(Matrix->Columns[j] >= CutRow)
				RightPortion ++;
		
		if(RightColumn < Matrix->Columns[Matrix->Pointers[i+1]-1])
			RightColumn = Matrix->Columns[Matrix->Pointers[i+1]-1];
	}
}

//NUMBER OF NONZEROS (BAND MATRIX)
void estimation(SparseMatrix *In, SparseMatrix *Out)
{
	int i, nb, row, col, j, index, decal, end, begin;

	index = offset % In->NumRows;
	end = In->NumRows - CutRow - 1;
	begin = In->NumRows - CutRow - 1;
	nb = 0;

	for(i=0; i<sizeloc; i++){
		decal = ((i + offset) / In->NumRows) * In->NumRows;
		
		//Left
		if((i+offset)>=In->NumRows && index>=0 && index<end){
			row = CutRow + index;
			for(j=In->Pointers[row]; j<In->Pointers[row+1]; j++){
				col = In->Columns[j] + decal - CutRow - 1;
				if(In->Columns[j]<=CutRow && col>=0 && col<sizeglob)
					nb ++;
			}
		}

		//Center
		row = index;
		for(j=In->Pointers[row]; j<In->Pointers[row+1]; j++){
			col = In->Columns[j] + decal;
			if(col>=0 && col<sizeglob)
				nb ++;
		}

		//Right
		if((i+offset)<(sizeglob-CutRow-1) && index>=begin && index<In->NumRows){
			row = index - begin;
			for(j=In->Pointers[row]; j<In->Pointers[row+1]; j++){	
				col = In->Columns[j] + decal + In->NumRows - CutRow;
				if(In->Columns[j]>=CutRow && col>=0 && col<sizeglob)
					nb ++; 	
			}
		}

		index ++;
		if(index == In->NumRows) index = 0;
	}

	Out->NumNonZeros = nb;	
}

void Generation_Band_Sparse_Matrix(SparseMatrix *In, SparseMatrix *Out)
{
	int i, nb, count, row, col, j, index, decal, end, begin;

	Out->Pointers[0] = 0;
	index = offset % In->NumRows;
	end = In->NumRows - CutRow - 1;
	begin = In->NumRows - CutRow - 1;
	nb = 0;

	for(i=0; i<sizeloc; i++){
		count = 0;
		decal = ((i + offset) / In->NumRows) * In->NumRows;
		
		//Left
		if((i+offset)>=In->NumRows && index>=0 && index<end){
			row = CutRow + index;
			for(j=In->Pointers[row]; j<In->Pointers[row+1]; j++){
				col = In->Columns[j] + decal - CutRow - 1;
				if(In->Columns[j]<=CutRow && col>=0 && col<sizeglob){
					Out->Columns[nb] = col;
					Out->NonZeros[nb] = In->NonZeros[j] / 100;
					nb ++;
					count ++; 
				}
			}
		}

		//Center
		row = index;
		for(j=In->Pointers[row]; j<In->Pointers[row+1]; j++){
			col = In->Columns[j] + decal;
			if(col>=0 && col<sizeglob){
				Out->Columns[nb] = col;
				Out->NonZeros[nb] = In->NonZeros[j]; 
				nb ++;
				count ++; 
			}
		}

		//Right
		if((i+offset)<(sizeglob-CutRow-1) && index>=begin && index<In->NumRows){
			row = index - begin;
			for(j=In->Pointers[row]; j<In->Pointers[row+1]; j++){	
				col = In->Columns[j] + decal + In->NumRows - CutRow;
				if(In->Columns[j]>=CutRow && col>=0 && col<sizeglob){
					Out->Columns[nb] = col;
					Out->NonZeros[nb] = In->NonZeros[j] / 100;
					nb ++;
					count ++; 
				}
	
			}
		}

		Out->Pointers[i+1] = Out->Pointers[i] + count;
		index ++;
		if(index == In->NumRows) index = 0;
	}

	Out->NumNonZeros = nb;

	free(In->Pointers);
	free(In->Columns);
	free(In->NonZeros);	
}

//NUMBER OF NONZEROS (3 BAND MATRIX & 2 PARTS)
void estimation_1(SparseMatrix *In, SparseMatrix *Out)
{
	int i, nb, col, j, index, decal, end, begin;
	int ColRow, ecart, indexL, decalL, decalR;

	end = In->NumRows - CutRow - 1;
	begin = In->NumRows - CutRow - 1;
	nb = 0;
	ecart = (sizeglob - 2 * In->NumRows) / 2;
	ColRow = In->NumRows + ecart;

	for(i=0; i<sizeloc; i++){
		decal = ((i + offset) / In->NumRows) * In->NumRows;
		decalL = ((i + offset - ColRow) / In->NumRows) * In->NumRows;
		decalR = ((i + offset) / In->NumRows) * In->NumRows + ColRow;

		//Left part
		indexL = ((i + offset) - (sizeglob - In->NumRows)) % In->NumRows;
		if((i+offset)>=In->NumRows && (i+offset)>=(sizeglob-In->NumRows)){
			for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
				col = In->Columns[j];
				if(col>=0 && col<In->NumRows && col>sizeglob){
					nb ++;
				}
			}
		}

		//Left (left band)
		indexL = ((i + offset) - ColRow) % In->NumRows;
		if((i+offset)>=(ColRow+In->NumRows) && indexL>=0 && indexL<end){
			for(j=In->Pointers[indexL+CutRow]; j<In->Pointers[indexL+CutRow+1]; j++){
				col = In->Columns[j] - CutRow - 1 + decalL;
				if(In->Columns[j]<=CutRow && col>=decalL-In->NumRows && col<decalL && col<sizeglob){
					if(!(i+offset>=sizeglob-In->NumRows && col<In->NumRows)){
						nb ++;
					}
				}
			}
		}

		//Center (left band)
		indexL = ((i + offset) - ColRow) % In->NumRows;
		if(i+offset>=ColRow){
			for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
				col = In->Columns[j] + decalL;
				if(col>=decalL && col<decalL+In->NumRows && col<sizeglob){
					if(!(col<In->NumRows && (i+offset<In->NumRows || i+offset>=sizeglob-In->NumRows))){
						nb ++; 
					}
				}
			}
		}

		//Right (left band)
		indexL = ((i + offset) - ColRow) % In->NumRows;
		if(i+offset>=ColRow && indexL>=begin && indexL<In->NumRows){
			for(j=In->Pointers[indexL-begin]; j<In->Pointers[indexL-begin+1]; j++){
				col = In->Columns[j] + In->NumRows + decalL - CutRow;
				if(In->Columns[j]>=CutRow && col>=In->NumRows+decalL && col<decal-CutRow-1 && col<sizeglob){
					nb ++; 
				}
			}
		}		

		//Left (main band)
		index = (i + offset) % In->NumRows;
		if((i+offset)>=In->NumRows && index>=0 && index<end){
			for(j=In->Pointers[index+CutRow]; j<In->Pointers[index+CutRow+1]; j++){
				col = In->Columns[j] + decal - CutRow - 1;
				if(In->Columns[j]<=CutRow && col>=decal-In->NumRows && col<decal && col<sizeglob){
					if(!(col<In->NumRows && (i+offset<In->NumRows || i+offset>=sizeglob-In->NumRows))){
						nb ++;
					}
				}
			}
		}

		//Center (main band)
		index = (i + offset) % In->NumRows;
		for(j=In->Pointers[index]; j<In->Pointers[index+1]; j++){
			col = In->Columns[j] + decal;
			if(col>=decal && col<sizeglob){
				nb ++; 
			}
		}

		//Right (main band)
		index = (i + offset) % In->NumRows;
		if((i+offset)<(sizeglob-CutRow-1) && index>=begin && index<In->NumRows){
			for(j=In->Pointers[index-begin]; j<In->Pointers[index-begin+1]; j++){	
				col = In->Columns[j] + decal + In->NumRows - CutRow;
				if(In->Columns[j]>=CutRow && col>=decal+In->NumRows && col<sizeglob){
					if(!((i+offset<In->NumRows && col>=sizeglob-In->NumRows) || col>=decalR)){
						nb ++;
					}
				}
	
			}
		}

		//Left (right band)
		indexL = (i + offset) % In->NumRows; 
		if((i+offset)>=In->NumRows && indexL>=0 && indexL<end){
			for(j=In->Pointers[indexL+CutRow]; j<In->Pointers[indexL+CutRow+1]; j++){
				col = In->Columns[j] + decalR - CutRow - 1;
				if(In->Columns[j]<=CutRow && col>=decalR-In->NumRows && col<decalR && col<sizeglob){
					if(!(col<decal+CutRow)){
						nb ++;
					}
				}
			}
		}

		//Center (right band)
		indexL = (i + offset) % In->NumRows; 
		for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
			col = In->Columns[j] + decalR;
			if(col>=ColRow && col<sizeglob){
				if(!(i+offset<In->NumRows && (col<In->NumRows || col>=sizeglob-In->NumRows))){
					nb ++;
				}
			} 
		}

		//Right (right band)
		indexL = (i + offset) % In->NumRows; 
		if(indexL>=begin && indexL<In->NumRows){
			for(j=In->Pointers[indexL-begin] ; j<In->Pointers[indexL-begin+1]; j++){
				col = In->Columns[j] - CutRow + decalR + In->NumRows;
				if(In->Columns[j]>=CutRow && col>=decalR+In->NumRows && col<sizeglob){
					if(!(i+offset<In->NumRows && col>=sizeglob-In->NumRows)){
						nb ++;
					}
				}
			}
		}

		//Right part
		indexL = (i + offset) % In->NumRows;
		if(i+offset<In->NumRows){
			for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
				col = In->Columns[j] + sizeglob - In->NumRows;
				if(col>=In->NumRows && col>=(sizeglob-In->NumRows) && col<sizeglob){
					nb ++;
				}
			}
		}
	}
	Out->NumNonZeros = nb;	
}

void Generation_Sparse_Matrix_Of_Three_Bands(SparseMatrix *In, SparseMatrix *Out)
{
	int i, nb, count, col, j, index, decal, end, begin;
	int ColRow, ecart, indexL, decalL, decalR;

	Out->Pointers[0] = 0;
	end = In->NumRows - CutRow - 1;
	begin = In->NumRows - CutRow - 1;
	nb = 0;
	ecart = (sizeglob - 2 * In->NumRows) / 2;
	ColRow = In->NumRows + ecart;

	for(i=0; i<sizeloc; i++){
		count = 0;
		decal = ((i + offset) / In->NumRows) * In->NumRows;
		decalL = ((i + offset - ColRow) / In->NumRows) * In->NumRows;
		decalR = ((i + offset) / In->NumRows) * In->NumRows + ColRow;

		//Left part
		indexL = ((i + offset) - (sizeglob - In->NumRows)) % In->NumRows;
		if((i+offset)>=In->NumRows && (i+offset)>=(sizeglob-In->NumRows)){
			for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
				col = In->Columns[j];
				if(col>=0 && col<In->NumRows && col>sizeglob){
					Out->Columns[nb] = col;
					Out->NonZeros[nb] = In->NonZeros[j] / 100;
					nb ++;
					count ++; 
				}
			}
		}

		//Left (left band)
		indexL = ((i + offset) - ColRow) % In->NumRows;
		if((i+offset)>=(ColRow+In->NumRows) && indexL>=0 && indexL<end){
			for(j=In->Pointers[indexL+CutRow]; j<In->Pointers[indexL+CutRow+1]; j++){
				col = In->Columns[j] - CutRow - 1 + decalL;
				if(In->Columns[j]<=CutRow && col>=decalL-In->NumRows && col<decalL && col<sizeglob){
					if(!(i+offset>=sizeglob-In->NumRows && col<In->NumRows)){
						Out->Columns[nb] = col;
						Out->NonZeros[nb] = In->NonZeros[j] / 100;
						nb ++;
						count ++; 
					}
				}
			}
		}

		//Center (left band)
		indexL = ((i + offset) - ColRow) % In->NumRows;
		if(i+offset>=ColRow){
			for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
				col = In->Columns[j] + decalL;
				if(col>=decalL && col<decalL+In->NumRows && col<sizeglob){
					if(!(col<In->NumRows && (i+offset<In->NumRows || i+offset>=sizeglob-In->NumRows))){
						Out->Columns[nb] = col;
						Out->NonZeros[nb] = In->NonZeros[j] / 100;
						nb ++;
						count ++; 
					}
				}
			}
		}

		//Right (left band)
		indexL = ((i + offset) - ColRow) % In->NumRows;
		if(i+offset>=ColRow && indexL>=begin && indexL<In->NumRows){
			for(j=In->Pointers[indexL-begin]; j<In->Pointers[indexL-begin+1]; j++){
				col = In->Columns[j] + In->NumRows + decalL - CutRow;
				if(In->Columns[j]>=CutRow && col>=In->NumRows+decalL && col<decal-CutRow-1 && col<sizeglob){
					Out->Columns[nb] = col;
					Out->NonZeros[nb] = In->NonZeros[j] / 100;
					nb ++;
					count ++; 
				}
			}
		}		

		//Left (main band)
		index = (i + offset) % In->NumRows;
		if((i+offset)>=In->NumRows && index>=0 && index<end){
			for(j=In->Pointers[index+CutRow]; j<In->Pointers[index+CutRow+1]; j++){
				col = In->Columns[j] + decal - CutRow - 1;
				if(In->Columns[j]<=CutRow && col>=decal-In->NumRows && col<decal && col<sizeglob){
					if(!(col<In->NumRows && (i+offset<In->NumRows || i+offset>=sizeglob-In->NumRows))){
						Out->Columns[nb] = col;
						Out->NonZeros[nb] = In->NonZeros[j] / 100;
						nb ++;
						count ++; 
					}
				}
			}
		}

		//Center (main band)
		index = (i + offset) % In->NumRows;
		for(j=In->Pointers[index]; j<In->Pointers[index+1]; j++){
			col = In->Columns[j] + decal;
			if(col>=decal && col<sizeglob){
				Out->Columns[nb] = col;
				Out->NonZeros[nb] = In->NonZeros[j] * 100; 
				nb ++;
				count ++; 
			}
		}

		//Right (main band)
		index = (i + offset) % In->NumRows;
		if((i+offset)<(sizeglob-CutRow-1) && index>=begin && index<In->NumRows){
			for(j=In->Pointers[index-begin]; j<In->Pointers[index-begin+1]; j++){	
				col = In->Columns[j] + decal + In->NumRows - CutRow;
				if(In->Columns[j]>=CutRow && col>=decal+In->NumRows && col<sizeglob){
					if(!((i+offset<In->NumRows && col>=sizeglob-In->NumRows) || col>=decalR)){
						Out->Columns[nb] = col;
						Out->NonZeros[nb] = In->NonZeros[j] / 100;
						nb ++;
						count ++; 
					}
				}
	
			}
		}

		//Left (right band)
		indexL = (i + offset) % In->NumRows; 
		if((i+offset)>=In->NumRows && indexL>=0 && indexL<end){
			for(j=In->Pointers[indexL+CutRow]; j<In->Pointers[indexL+CutRow+1]; j++){
				col = In->Columns[j] + decalR - CutRow - 1;
				if(In->Columns[j]<=CutRow && col>=decalR-In->NumRows && col<decalR && col<sizeglob){
					if(!(col<decal+CutRow)){
						Out->Columns[nb] = col;
						Out->NonZeros[nb] = In->NonZeros[j] / 100;
						nb ++;
						count ++;
					}
				}
			}
		}

		//Center (right band)
		indexL = (i + offset) % In->NumRows; 
		for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
			col = In->Columns[j] + decalR;
			if(col>=ColRow && col<sizeglob){
				if(!(i+offset<In->NumRows && (col<In->NumRows || col>=sizeglob-In->NumRows))){
					Out->Columns[nb] = col;
					Out->NonZeros[nb] = In->NonZeros[j] / 100;
					nb ++;
					count ++;
				}
			} 
		}
			
		//Right (right band)
		indexL = (i + offset) % In->NumRows; 
		if(indexL>=begin && indexL<In->NumRows){
			for(j=In->Pointers[indexL-begin] ; j<In->Pointers[indexL-begin+1]; j++){
				col = In->Columns[j] - CutRow + decalR + In->NumRows;
				if(In->Columns[j]>=CutRow && col>=decalR+In->NumRows && col<sizeglob){
					if(!(i+offset<In->NumRows && col>=sizeglob-In->NumRows)){
						Out->Columns[nb] = col;
						Out->NonZeros[nb] = In->NonZeros[j] / 100;
						nb ++;
						count ++;
					}
				}
			}
		}

		//Right Part
		indexL = (i + offset) % In->NumRows;
		if(i+offset<In->NumRows){
			for(j=In->Pointers[indexL]; j<In->Pointers[indexL+1]; j++){
				col = In->Columns[j] + sizeglob - In->NumRows;
				if(col>=In->NumRows && col>=(sizeglob-In->NumRows) && col<sizeglob){
					Out->Columns[nb] = col;
					Out->NonZeros[nb] = In->NonZeros[j] / 100;
					nb ++;
					count ++;
				}
			}
		}

		Out->Pointers[i+1] = Out->Pointers[i] + count;
	}

	Out->NumNonZeros = nb;

	free(In->Pointers);
	free(In->Columns);
	free(In->NonZeros);	
}

//Compute the ends of the matrix
void Computing_Borders(SparseMatrix *matrix)
{
	int i;
	
	LeftColumn = matrix->NumColumns - 1;
	RightColumn = 0;
	
	for(i=0; i<matrix->NumRows; i++){
		if(LeftColumn > matrix->Columns[matrix->Pointers[i]])
			LeftColumn = matrix->Columns[matrix->Pointers[i]];
		
		
		if(RightColumn < matrix->Columns[matrix->Pointers[i+1]-1])
			RightColumn = matrix->Columns[matrix->Pointers[i+1]-1];
	}
}

//CONVERSION FROM CSC TO CSR & VICE VERSA
void Conversion_CSC_CSR(SparseMatrix *In, SparseMatrix *Out)
{
	int i, j, col, index;
	int *vect;

//INITIALIZATION
	Out->NumRows = In->NumColumns;
	Out->NumColumns = In->NumRows;
	Out->NumNonZeros = In->NumNonZeros;

	Out->Pointers = (int*)malloc(sizeof(int)*(Out->NumRows+1));
	Out->Columns = (int*)malloc(sizeof(int)*Out->NumNonZeros);
	Out->NonZeros = (double*)malloc(sizeof(double)*Out->NumNonZeros);

//COMPUTE POINTERS	
	vect = (int*)malloc(sizeof(int)*(Out->NumRows+1));
	for(i=0; i<Out->NumRows+1; i++) vect[i] = 0;
	for(i=0; i<In->NumNonZeros; i++) vect[In->Columns[i]] ++;

	Out->Pointers[0] = 0;
	for(i=0; i<Out->NumRows; i++){
		Out->Pointers[i+1] = Out->Pointers[i] + vect[i];
		vect[i] = Out->Pointers[i];
	}
	vect[Out->NumRows] = Out->Pointers[Out->NumRows];

//COMPUTE COLUMNS & NONZEROS
	for(i=0; i<In->NumRows; i++){
		for(j=In->Pointers[i]; j<In->Pointers[i+1]; j++){
			col = In->Columns[j];
			index = vect[col];
			Out->NonZeros[index] = In->NonZeros[j];
			Out->Columns[index] = i;
			vect[col] ++;		
		}
	}

//CLEAN
	free(vect);	
	free(In->Pointers);
	free(In->Columns);
	free(In->NonZeros);
}

//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          CPU FUNCTIONS          ----------------------------------------------------------//
void init_values(int size, int columns, int restart, double *h_x, double *h_b, double *h_r, double *h_s, double *h_tmp_V)
{
	std::fill(h_s, h_s+restart+1, 0.0);

	for(int i=0; i<columns; i++){
		h_x[i] = 0.0;
		h_tmp_V[i] = 0.0;
	}
	
	for(int i=0; i<size; i++){
		h_b[i] = 1.0;
		h_r[i] = 0.0;
	}
}

void csrmul(double *y, int *Ap,  int *Aj, double *Av, int num_rows, double *x)
{
	int row,col;
	double sum;

	for(row=0; row<num_rows; row++){
		sum = 0;
		for(col=Ap[row]; col<Ap[row+1]; col++) 
			sum += Av[col] * x[Aj[col]];		
		y[row] = sum; 
	}
}

void yminusx(int n, double *x, double *y)
{
	int i;
	for(i=0; i<n; i++)
		y[i] = x[i] - y[i];
}

void diag_solve(int size, double *h_d, double *h_b, double* h_tmp)
{
	int i;
  	for(i=0; i<size; i++)
		h_tmp[i] = h_d[i] * h_b[i];
}

double norm(int size, double *vect)
{
	double dum = 0;
	int i;
	for(i=0; i<size; i++) {
		dum += vect[i] * vect[i];
	}
	return sqrt(dum);
}

double vect_dot(int size, double *vecta, double *vectb) 
{
	double ret = 0;
  	int i;
  	for(i=0; i<size; i++) {
    		ret += vecta[i] * vectb[i];
  	}
  	return ret;
}

void ApplyPlaneRotation(double *dx, double *dy, double *cs, double *sn)
{
  double temp  =  *cs * *dx + *sn * *dy;
  *dy = -*sn * *dx + *cs * *dy;
  *dx = temp;
}

void GeneratePlaneRotation(double *dx, double *dy, double *cs, double *sn)
{
  if (*dy == 0.0) {
    *cs = 1.0;
    *sn = 0.0;
  } else if (fabs(*dy) > fabs(*dx)) {
    double temp = *dx / *dy;
    *sn = 1.0 / sqrt( 1.0 + temp*temp );
    *cs = temp * *sn;
  } else {
    double temp = *dy / *dx;
    *cs = 1.0 / sqrt( 1.0 + temp*temp );
    *sn = temp * *cs;
  }
}

void Update(int size2, double *x, int k, double *H, double *s, double *v)
{
	int i,j,kk;
	double y[restart];
	
	for(kk=0; kk<restart; kk++)
		y[kk] = s[kk];
  	// Backsolve:  
  	for (i=k; i>=0; i--){
    		y[i] /= H[i*restart+i];
		for (j=i-1; j>=0; j--) 
      			y[j] -= H[j*restart+i] * y[i];
	}
	for(j=0; j<=k; j++)
		for(kk=0; kk<size2; kk++) 
			x[kk] += v[j*sizeloc+kk] * y[j];
}

void Compress_Vector(SparseMatrix *Matrix, double* h_x)
{
	int i;
	for(i=0; i<Matrix->send; i++)
		Matrix->CompressedX[i] = h_x[Matrix->SendColumns[i]];
}


void Synchronize(SparseMatrix *Matrix, double *vect)
{
	MPI_Status st;
	int i, d, RANK;

	for(i=0; i<Matrix->ranksend; i++){//send
		RANK = Matrix->RankSend[i];
		d = Matrix->sdispls[RANK];
		MPI_Issend(&Matrix->CompressedX[d], Matrix->sendcount[RANK], MPI_DOUBLE, RANK, Matrix->tag, MPI_COMM_WORLD, &Matrix->SendReq[RANK]);
	}

	for(i=0; i<Matrix->rankrecv; i++){//recv
		RANK = Matrix->RankRecv[i];
		d = Matrix->rdispls[RANK];
		MPI_Irecv(&vect[d], Matrix->recvcount[RANK], MPI_DOUBLE, RANK, Matrix->tag, MPI_COMM_WORLD, &Matrix->RecvReq[RANK]);
	}

	for(i=0; i<Matrix->ranksend; i++){//wait for send
		RANK = Matrix->RankSend[i];
		MPI_Wait(&Matrix->SendReq[RANK], &st);
	}

	for(i=0; i<Matrix->rankrecv; i++){//wait for recv
		RANK = Matrix->RankRecv[i];
		MPI_Wait(&Matrix->RecvReq[RANK], &st);
	}  	
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//


//---------------------------------------------------------          CPU GMRES          --------------------------------------------------------------//
void GMRES_On_CPU(double *gmres_cpu_timer, double *h_x_gmres_cpu,
		  double *h_b, double *h_r, double *h_d, double *h_H, double *h_v, double *h_s, double *h_cs,
		  double *h_sn, double *h_w, double *h_tmp_V, double *h_tmp_v1, double *h_tmp_v2)
{
	double t1, t2, normb, beta, val, valeur, Max, time_cpu, TIME_CPU;
	int i, j, k, found, iter, ii;

	if(rank == rang){
		printf("\n\n\t\t\t\t ******   START OF CPU COMPUTATION   ******");
		printf("\n\t\t\t\t ******   START OF CPU COMPUTATION   ******\n");
	}

	//Initialisation
	init_values(sizeloc, XSize, restart, h_x_gmres_cpu, h_b, h_r, h_s, h_tmp_V);
	MPI_Barrier(MPI_COMM_WORLD);

	//GMRES on the CPU cluster
	t1 = MPI_Wtime();
	csrmul(h_r, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_x_gmres_cpu);
	yminusx(sizeloc, h_b, h_r);
	diag_solve(sizeloc, h_d, h_r, h_r);
	diag_solve(sizeloc, h_d, h_b, h_w);
	normb = norm(sizeloc, h_w);
	beta = norm(sizeloc, h_r);
	h_tmp_v1[0] = normb * normb;
	h_tmp_v1[1] = beta * beta;
	MPI_Allreduce(h_tmp_v1, h_tmp_v2, 2, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	normb = sqrt(h_tmp_v2[0]);
	beta = sqrt(h_tmp_v2[1]);
	if(rank == rang){
		printf("\n\t\t -- CPU normb_begin : %20.20e", normb);
		printf("\n\t\t -- CPU beta_begin : %20.20e", beta);
	}
	j = 1;
  	found = 0;

 	while(beta>precision && j<maxiter && found==0){
    		for(ii=0; ii<sizeloc; ii++)
      			h_v[ii] = h_r[ii] / beta;
       		for(ii=0; ii<restart+1; ii++)
      			h_s[ii] = 0.0;
    		h_s[0] = beta;
   		for(i=0; i<restart && j<maxiter && found==0; i++, j++){
			for(ii=0; ii<sizeloc; ii++)
	 			h_tmp_V[ii] = h_v[i*sizeloc+ii];
			Compress_Vector(&Matrix, h_tmp_V);
			Matrix.tag = 100;
			Synchronize(&Matrix, &h_tmp_V[sizeloc]);
			csrmul(h_w, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_tmp_V);
			diag_solve(sizeloc, h_d, h_w, h_w);
			for(k=0; k<=i; k++)
	 			 h_tmp_v1[k] = vect_dot(sizeloc, h_w, &h_v[k*sizeloc]);
			MPI_Allreduce(h_tmp_v1, h_tmp_v2, i+1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			for(k=0; k<=i; k++)
	  			h_H[k*restart+i] = h_tmp_v2[k];		
			for(k=0; k<=i; k++) 
	  			for(ii=0; ii<sizeloc; ii++) 
	    				h_w[ii] -= h_H[k*restart+i] * h_v[k*sizeloc+ii];
	  		val = norm(sizeloc, h_w);
	   		val *= val;
			MPI_Allreduce(&val, &h_H[(i+1)*restart+i], 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			h_H[(i+1)*restart+i] = sqrt(h_H[(i+1)*restart+i]);
			for(ii=0; ii<sizeloc; ii++) 
	  			h_v[(i+1)*sizeloc+ii] = h_w[ii]/h_H[(i+1)*restart+i];
			for (k=0; k<i; k++)
	  			ApplyPlaneRotation(&h_H[k*restart+i], &h_H[(k+1)*restart+i], &h_cs[k], &h_sn[k]);
			GeneratePlaneRotation(&h_H[i*restart+i], &h_H[(i+1)*restart+i], &h_cs[i], &h_sn[i]);
			ApplyPlaneRotation(&h_H[i*restart+i], &h_H[(i+1)*restart+i], &h_cs[i], &h_sn[i]);
			ApplyPlaneRotation(&h_s[i], &h_s[i+1], &h_cs[i], &h_sn[i]);			
			if(fabs(h_s[i+1])/normb < precision){
	  			Update(sizeloc, h_x_gmres_cpu, i, h_H, h_s, h_v);
				found = 1;				
			}
		}
		if(found==0){
      			Update(sizeloc, h_x_gmres_cpu, restart-1, h_H, h_s, h_v);
			Compress_Vector(&Matrix, h_x_gmres_cpu);
			Matrix.tag = 101;
			Synchronize(&Matrix, &h_x_gmres_cpu[sizeloc]);
			csrmul(h_w, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_x_gmres_cpu);
      			yminusx(sizeloc, h_b, h_w);
      			diag_solve(sizeloc, h_d, h_w, h_r);
		        beta = norm(sizeloc, h_r);
      			beta *= beta;
      			MPI_Allreduce(&beta, &valeur, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
      			beta = sqrt(valeur);
      			if(beta/normb < precision)
				found = 1;
		}
	}
	MPI_Barrier(MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	time_cpu = (t2 - t1) * 1000;
	MPI_Allreduce(&time_cpu, &TIME_CPU, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	*gmres_cpu_timer = TIME_CPU / nbprocs;
	iter = j;
	if(rank == rang){
		printf("\n\t\t -- CPU processing time : %f (s)", *gmres_cpu_timer/1000);
		printf("\n\t\t -- CPU iter_end : %d", iter);
		printf("\n\t\t -- CPU normb_end : %20.20e", normb);
		printf("\n\t\t -- CPU beta_end : %20.20e", beta);
	}
	for(i=0; i<sizeloc; i++) h_r[i] = 0.0;
	Compress_Vector(&Matrix, h_x_gmres_cpu);
	Matrix.tag = 102;
	Synchronize(&Matrix, &h_x_gmres_cpu[sizeloc]);
	csrmul(h_r, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_x_gmres_cpu);
	yminusx(sizeloc, h_b, h_r);
	diag_solve(sizeloc, h_d, h_r, h_r);

	Max = 0.0; valeur = 0.0;
	for(i=0; i<sizeloc; i++)
    		if(Max<fabs(h_r[i]))
      			Max = fabs(h_r[i]); 
	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
	if(rank == rang)
		printf("\n\t\t -- CPU MAX of residu : %e\n\n", valeur);
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          GPU FUNCTIONS          ----------------------------------------------------------//
static __inline__ __device__ double fetch_double(texture<int2,1> t, int i)
{
        int2 v = tex1Dfetch(t, i);
        return __hiloint2double(v.y, v.x);
}

__global__
void yminusx_kernel(int n, double *x, double *y)
{
	int i = blockIdx.x * blockDim.x + threadIdx.x;
	if(i < n)   
		y[i] = x[i] - y[i];
}

__global__
void diag_solve_kernel(int size, double *d, double *b, double* tmp)
{
	int i = blockIdx.x * blockDim.x + threadIdx.x;
  	if(i < size)
    		tmp[i] = d[i] * b[i];
}

__global__
void scale_kernel(int n, double* v, double *r, double inv_beta)
{
  	int i = blockIdx.x * blockDim.x + threadIdx.x;
  	if(i < n)
		v[i] = r[i] * inv_beta;
}

__global__
void s_kernel(int r,double *s)
{
  	int i = blockIdx.x * blockDim.x + threadIdx.x;
  	if(i < r)
    		s[i] = 0.0;
}

__device__
void ApplyPlaneRotation2(double *dx, double *dy, double *cs, double *sn)
{
  	double temp  =  *cs * *dx + *sn * *dy;
  	*dy = -*sn * *dx + *cs * *dy;
  	*dx = temp;
}

__device__
void GeneratePlaneRotation2(double *dx, double *dy, double *cs, double *sn)
{
	if(*dy == 0.0){
	    	*cs = 1.0;
	    	*sn = 0.0;
	}
	else{ 
		if(fabs(*dy) > fabs(*dx)){
			double temp = *dx / *dy;
			*sn = 1.0 / sqrt(1.0 + temp * temp);
			*cs = temp * *sn;
		}
		else{
			double temp = *dy / *dx;
			*cs = 1.0 / sqrt(1.0 + temp * temp);
			*sn = temp * *cs;
  		}
	}
}

__global__
void ApplyPlaneRotation_kernel(int i, int restart, double *h_H, double *h_cs, double *h_sn, double *h_s)
{
//EXECTUTED BY ONE THREAD
	int kk;
	for (kk=0; kk<i; kk++)
		ApplyPlaneRotation2(&h_H[kk*restart+i], &h_H[(kk+1)*restart+i], &h_cs[kk], &h_sn[kk]);

	GeneratePlaneRotation2(&h_H[i*restart+i], &h_H[(i+1)*restart+i], &h_cs[i], &h_sn[i]);
	ApplyPlaneRotation2(&h_H[i*restart+i], &h_H[(i+1)*restart+i], &h_cs[i], &h_sn[i]);
	ApplyPlaneRotation2(&h_s[i], &h_s[i+1], &h_cs[i], &h_sn[i]);
}

__global__
void Update_kernel(int restart, int k, double *H, double *s, double *y)
{
//EXECTUTED BY ONE THREAD
  	int i, j;
	// Backsolve:
	for (i = k; i >= 0; i--){
		y[i] /= H[i*restart+i];
      		for (j = i - 1; j >= 0; j--)
			y[j] -= H[j*restart+i] * y[i];
	}
}

__global__
void Update_kernel2(int sizeloc, double *x, int k, double *v, double *y)
{
	int t = blockIdx.x * blockDim.x + threadIdx.x;
  	if(t < sizeloc){
    		for (int j = 0; j <= k; j++)
      			x[t] += v[j*sizeloc+t] * y[j];
	}
}

__global__
void Device_Compress_Vector(int send, double* d_x, double* d_Send_Compressed_X, int* d_Send_Columns)
{
	int t = blockIdx.x * blockDim.x + threadIdx.x;
  	if(t < send)
		d_Send_Compressed_X[t] = /*fetch_double(TextureX, d_Send_Columns[t]);*/d_x[d_Send_Columns[t]];
	
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          GPU GMRES          --------------------------------------------------------------//
void GMRES_On_GPU(csr_matrix<int,double> csr, double **h_x_gmres_gpu, double *gpu_timer, double *h_x_gmres_cpu,
		  double *h_b, double *h_r, double *h_d, double *h_H, double *h_v, double *h_s, double *h_cs,
		  double *h_sn, double *h_w, double *h_tmp_V, double *h_tmp_v1, double *h_tmp_v2) 
{
	double normb, beta, val, Max, valeur, time_gpu, TIME_GPU;
	int i, k, j, iter, found;
	int threads = 512;
	unsigned int timer_proc = 0;
	int send = Matrix.send;
	int recv = Matrix.recv;
	int blocks = (sizeloc + threads - 1) / threads;
	int threads_send = 512;
	int blocks_send = (send + threads_send - 1) / threads_send;
	
	//Initialisation
	init_values(sizeloc, XSize, restart, h_x_gmres_cpu, h_b, h_r, h_s, h_tmp_V);
	MPI_Barrier(MPI_COMM_WORLD);
		
	//Matrix conversion from csr to hyp
	const int num_cols_per_row = compute_hyb_cols_per_row(csr);
	hyb_matrix<int,double> hyb = csr_to_hyb<int,double>(csr, num_cols_per_row);
	if(rank == rang)
		printf("\n\t\t -- ELL has : %d columns : \n\t\t -- NNZ : [ELL %4.1f%%] et [COO %4.1f%%]\n", num_cols_per_row,
				 100 * (double) hyb.ell.num_nonzeros / (double) csr.num_nonzeros, 
				 100 * (double) hyb.coo.num_nonzeros / (double) csr.num_nonzeros);
	
	const memory_location loc = DEVICE_MEMORY;
	hyb_matrix<int,double> sm_loc = (loc == DEVICE_MEMORY) ? copy_matrix_to_device(hyb) : hyb;
	
	// create vectors in appropriate locations
	double* y_host = new_host_array<double>(sizeloc);   
	double* decompressedX = new_host_array<double>(recv);   
	for(i=0; i<sizeloc; i++) y_host[i] = 0.0;
	for(i=0; i<send; i++) Matrix.CompressedX[i] = 0.0;
	for(i=0; i<recv; i++) decompressedX[i] = 0.0;

	double * d_x = copy_array(h_x_gmres_cpu, XSize, HOST_MEMORY, loc);
	double * d_b = copy_array(h_b, sizeloc, HOST_MEMORY, loc);
	double * d_d = copy_array(h_d, sizeloc, HOST_MEMORY, loc);
	double * d_w = copy_array(y_host, sizeloc, HOST_MEMORY, loc);       
	double * d_r = copy_array(y_host, sizeloc, HOST_MEMORY, loc);       
	double * d_v = new_device_array<double>(sizeloc*(restart+1));
	double * d_s = new_device_array<double>((restart+1));
	double * d_cs = new_device_array<double>((restart+1));
	double * d_sn = new_device_array<double>((restart+1));
	double * d_y = new_device_array<double>((restart+1));
	double * d_H = new_device_array<double>((restart+1)*restart);
	double * d_tmp_V = copy_array(h_tmp_V, XSize, HOST_MEMORY, loc);
	double * d_compressedX = copy_array(Matrix.CompressedX, send, HOST_MEMORY, loc);
	int * d_SendColumns = copy_array(Matrix.SendColumns, send, HOST_MEMORY, loc);
	free(y_host);
			
	// Start of computation
	if(rank == rang){
		printf("\n\t\t\t\t ******   START OF GPU COMPUTATION   ******");
		printf("\n\t\t\t\t ******   START OF GPU COMPUTATION   ******\n");
	}
	CUDA_SAFE_CALL(cudaThreadSynchronize());
	MPI_Barrier(MPI_COMM_WORLD);
	CUT_SAFE_CALL(cutCreateTimer(&timer_proc));
  	CUT_SAFE_CALL(cutStartTimer(timer_proc));
	spmv_hyb_tex_device<int,double>(sm_loc, d_x, d_r);
	yminusx_kernel<<<blocks,threads>>>(sizeloc, d_b, d_r);
	diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_r, d_r);
	diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_b, d_w);
	normb = cublasDnrm2(sizeloc, d_w, 1);
	beta = cublasDnrm2(sizeloc, d_r, 1);
	h_tmp_v1[0] = normb * normb;
  	h_tmp_v1[1] = beta * beta;
  	MPI_Allreduce(h_tmp_v1, h_tmp_v2, 2, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  	normb = sqrt(h_tmp_v2[0]);
  	beta = sqrt(h_tmp_v2[1]);
	if(rank == rang){
		printf("\n\t\t -- GPU normb_begin : %20.20e", normb);
		printf("\n\t\t -- GPU beta_begin : %20.20e", beta);
	}
	j = 1;
	found = 0;
	while(beta>precision && j<maxiter && found==0){
		scale_kernel<<<blocks,threads>>>(sizeloc, d_v, d_r, 1.0/beta);
		s_kernel<<<blocks,threads>>>(restart+1, d_s);
		cublasSetVector(1, sizeof(double), &beta, 1, &d_s[0], 1);
		for(i=0; i<restart && j<maxiter && found==0; i++, j++){	
			s_kernel<<<blocks,threads>>>(sizeloc, d_w);
			cublasDcopy(sizeloc, &d_v[i*sizeloc], 1, d_tmp_V, 1);
			//CUDA_SAFE_CALL(cudaBindTexture(NULL, TextureX, d_tmp_V)); 
			Device_Compress_Vector<<<blocks_send,threads_send>>>(send, d_tmp_V, d_compressedX, d_SendColumns);
			//CUDA_SAFE_CALL(cudaUnbindTexture(TextureX)); 
			cublasGetVector(send, sizeof(double), d_compressedX, 1, Matrix.CompressedX, 1);
			Matrix.tag = 100;
			Synchronize(&Matrix, decompressedX);
			cublasSetVector(recv, sizeof(double), decompressedX, 1, &d_tmp_V[sizeloc], 1);
			spmv_hyb_tex_device<int,double>(sm_loc, d_tmp_V, d_w);
			diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_w, d_w);
			for(k=0; k<=i; k++)
				h_tmp_v1[k] = cublasDdot(sizeloc, d_w, 1, &d_v[k*sizeloc], 1);
			MPI_Allreduce(h_tmp_v1, h_tmp_v2, i+1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			cublasSetVector(i+1, sizeof(double), h_tmp_v2, 1, &d_H[i], restart);
			for(k=0; k<=i; k++)
				cublasDaxpy(sizeloc, -h_tmp_v2[k], &d_v[k*sizeloc], 1, d_w, 1 );
			val = cublasDnrm2(sizeloc, d_w, 1);
			val *= val;
			MPI_Allreduce(&val, &valeur, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			val = sqrt(valeur);
			cublasSetVector(1, sizeof(double), &val, 1, &d_H[(i+1)*restart+i], 1);
			scale_kernel<<<blocks,threads>>>(sizeloc, &d_v[(i+1)*sizeloc], d_w, 1.0/val);
			ApplyPlaneRotation_kernel<<<1,1>>>(i, restart, d_H, d_cs, d_sn, d_s);
			cublasGetVector(1, sizeof(double), &d_s[i+1], 1, &val, 1);						
			if(fabs(val)/normb < precision){
				cublasDcopy(restart, d_s, 1, d_y, 1);
				Update_kernel<<<1,1>>>(restart, i, d_H, d_s, d_y);
				Update_kernel2<<<blocks,threads>>>(sizeloc, d_x, i, d_v, d_y);
				found = 1;				
			}
		}
		if(found == 0){
			cublasDcopy(restart, d_s, 1, d_y, 1);
			Update_kernel<<<1,1>>>(restart, restart-1, d_H, d_s, d_y);
			Update_kernel2<<<blocks,threads>>>(sizeloc, d_x, restart-1, d_v, d_y);
			s_kernel<<<blocks,threads>>>(sizeloc, d_w);
			//CUDA_SAFE_CALL(cudaBindTexture(NULL, TextureX, d_x)); 
			Device_Compress_Vector<<<blocks_send,threads_send>>>(send, d_x, d_compressedX, d_SendColumns);
			//CUDA_SAFE_CALL(cudaUnbindTexture(TextureX)); 
			cublasGetVector(send, sizeof(double), d_compressedX, 1, Matrix.CompressedX, 1);
			Matrix.tag = 101;
			Synchronize(&Matrix, decompressedX);
			cublasSetVector(recv, sizeof(double), decompressedX, 1, &d_x[sizeloc], 1);
			spmv_hyb_tex_device<int,double>(sm_loc, d_x, d_w);
			yminusx_kernel<<<blocks,threads>>>(sizeloc, d_b, d_w);
			diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_w, d_r);
			beta = cublasDnrm2(sizeloc, d_r, 1);
			beta *= beta;
			MPI_Allreduce(&beta, &valeur,1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			beta = sqrt(valeur);
			if(beta/normb < precision) 
				found = 1;
		}
	}
	CUT_CHECK_ERROR("Kernel execution failed");
  	CUDA_SAFE_CALL(cudaThreadSynchronize());
	MPI_Barrier(MPI_COMM_WORLD);
	cutStopTimer(timer_proc);
	time_gpu = cutGetTimerValue(timer_proc);
	cutDeleteTimer(timer_proc);
	MPI_Allreduce(&time_gpu, &TIME_GPU, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	iter = j;
	iteration = j;
	*gpu_timer = TIME_GPU / nbprocs;
	if(rank == rang){
		printf("\n\t\t -- GPU processing time : %f (s)", *gpu_timer/1000);
		printf("\n\t\t -- GPU iter_end : %d", iter);
		printf("\n\t\t -- GPU normb_end : %20.20e", normb);
		printf("\n\t\t -- GPU beta_end : %20.20e", beta);
	}
		
	s_kernel<<<blocks,threads>>>(sizeloc, d_r);
	//CUDA_SAFE_CALL(cudaBindTexture(NULL, TextureX, d_x)); 
	Device_Compress_Vector<<<blocks_send,threads_send>>>(send, d_x, d_compressedX, d_SendColumns);
	//CUDA_SAFE_CALL(cudaUnbindTexture(TextureX)); 
	cublasGetVector(send, sizeof(double), d_compressedX, 1, Matrix.CompressedX, 1);
	Matrix.tag = 102;
	Synchronize(&Matrix, decompressedX);
	cublasSetVector(recv, sizeof(double), decompressedX, 1, &d_x[sizeloc], 1);
	spmv_hyb_tex_device<int,double>(sm_loc, d_x, d_r);	
	yminusx_kernel<<<blocks,threads>>>(sizeloc, d_b, d_r);
	diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_r, d_r);
	cublasGetVector(sizeloc, sizeof(double), d_r, 1, h_r, 1); 
	Max = 0.0; 
	for(i=0; i<sizeloc; i++)
    		if(Max<fabs(h_r[i]))
      			Max = fabs(h_r[i]); 
				 
  	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
	if(rank == rang)
		printf("\n\t\t -- GPU MAX of residu : %e\n\n", valeur);
	*h_x_gmres_gpu = copy_array(d_x, sizeloc, loc, HOST_MEMORY);
		 
	//clean up device memory
	if(loc==DEVICE_MEMORY) delete_device_matrix(sm_loc);

	delete_array(d_x, loc);
	delete_array(d_b, loc);
	delete_array(d_d, loc);
	delete_array(d_r, loc);
	delete_array(d_w, loc);
	delete_array(d_v, loc);
	delete_array(d_s, loc);
	delete_array(d_cs, loc);
	delete_array(d_sn, loc);
	delete_array(d_y, loc);
	delete_array(d_H, loc);
	delete_array(d_tmp_V, loc);
	free(decompressedX);
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          CPU FUNCTIONS          ----------------------------------------------------------//
void init(int Size, int XSize, double *x, double *b, double *p, double *q, double *r, double *z)
{
	int i;
	
	for(i=0; i<Size; i++){
		b[i] = 1.0;
		q[i] = 0.0;
		r[i] = 0.0;
		z[i] = 0.0;
	}

	for(i=0; i<XSize; i++){
		x[i] = 0.0;
		p[i] = 0.0;
	}
}

void comp_norm(int Size, double *norm, double *x)
{
	int i;
	*norm = 0.0;
	for(i=0; i<Size; i++)
		if(fabs(x[i]) > *norm)
			*norm = fabs(x[i]);
}

void xequaly(int Size, double *x, double *y)
{
	int i;

	for(i=0; i<Size; i++)
		x[i] = y[i];
}

void xeaxpy(int Size, double a, double *x, double *y)
{
	int i;

	for(i=0; i<Size; i++)
		x[i] = a * x[i] + y[i];
}

void yeaxpy(int Size, double a, double *x, double *y)
{
	int i;

	for(i=0; i<Size; i++)
		y[i] = a * x[i] + y[i];
}

void yeymax(int Size, double a, double *x, double *y)
{
	int i;

	for(i=0; i<Size; i++)
		y[i] = y[i] - a * x[i];
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          CPU GMRES          --------------------------------------------------------------//
void CG_On_CPU(double *cg_cpu_timer, double *h_x_cg_cpu,
	       double *h_d, double *h_b, double *h_r, double *h_p, double *h_q, double *h_z)
{
	double t1, t2, time_cpu, TIME_CPU, normr, valeur, val, norm_algo, rho, rho_old, beta, alpha, Max;
	int iter, i;

	if(rank == rang){
		printf("\n\n\t\t\t\t ******   START OF CPU COMPUTATION   ******");
		printf("\n\t\t\t\t ******   START OF CPU COMPUTATION   ******\n");
	}

	//Initialisation
	init(sizeloc, XSize, h_x_cg_cpu, h_b, h_p, h_q, h_r, h_z);
	MPI_Barrier(MPI_COMM_WORLD);
	
	//CG on the CPU cluster
	t1 = MPI_Wtime();
	csrmul(h_r, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_x_cg_cpu);
	yminusx(sizeloc, h_b, h_r);
	comp_norm(sizeloc, &normr, h_r);
	MPI_Allreduce(&normr, &norm_algo, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

	if(rank == rang)
		printf("\n\t\t -- CPU norm_begin : %20.20e", norm_algo);
	iter = 0;
		
	while(norm_algo>precision && iter<maxiter){
		diag_solve(sizeloc, h_d, h_r, h_z);
		valeur = vect_dot(sizeloc, h_r, h_z);
		MPI_Allreduce(&valeur, &rho, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			
		if(iter == 0)
			xequaly(sizeloc, h_p, h_z);
		else{
			beta = rho / rho_old;
			xeaxpy(sizeloc, beta, h_p, h_z);
		}	

		Compress_Vector(&Matrix, h_p);
		Matrix.tag = 100;
		Synchronize(&Matrix, &h_p[sizeloc]);	
		csrmul(h_q, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_p);

		valeur = vect_dot(sizeloc, h_p, h_q);
		MPI_Allreduce(&valeur, &val, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

		alpha = rho / val;
		yeaxpy(sizeloc, alpha, h_p, h_x_cg_cpu);
		yeymax(sizeloc, alpha, h_q, h_r);
		rho_old = rho;
		iter ++;
		norm_algo = rho_old;
	}    		
	MPI_Barrier(MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	time_cpu = (t2 - t1) * 1000;
	MPI_Allreduce(&time_cpu, &TIME_CPU, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	*cg_cpu_timer = TIME_CPU / nbprocs;
	

	for(i=0; i<sizeloc; i++) h_r[i] = 0.0;
	Compress_Vector(&Matrix, h_x_cg_cpu);
	Matrix.tag = 101;
	Synchronize(&Matrix, &h_x_cg_cpu[sizeloc]);
	csrmul(h_r, Matrix.Pointers, Matrix.Columns, Matrix.NonZeros, sizeloc, h_x_cg_cpu);
	yminusx(sizeloc, h_b, h_r);
	diag_solve(sizeloc, h_d, h_r, h_r);

	Max = 0.0; valeur = 0.0;
	for(i=0; i<sizeloc; i++)
    		if(Max < fabs(h_r[i]))
      			Max = fabs(h_r[i]); 
	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

	if(rank == rang){
		printf("\n\t\t -- CPU processing time : %f (s)", *cg_cpu_timer/1000);
		printf("\n\t\t -- CPU iter_end : %d", iter);
		printf("\n\t\t -- CPU normb_end : %20.20e", norm_algo);
		printf("\n\t\t -- CPU MAX of residu : %e\n\n", valeur);
	}
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          GPU KERNELS          --------------------------------------------------------------//
__global__
void device_xeaxpy(int Size, double a, double *x, double *y)
{
	int t = blockIdx.x * blockDim.x + threadIdx.x;
	
	if(t < Size)
		x[t] = a * x[t] + y[t];
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------          GPU CG          -----------------------------------------------------------------//
void CG_On_GPU(csr_matrix<int,double> csr, double **h_x_cg_gpu, double *cg_gpu_timer, double *h_x_cg_cpu, 
	       double *h_d, double *h_b, double *h_r, double *h_p, double *h_q, double *h_z) 
{
	double time_gpu, TIME_GPU, normr, valeur, val, norm_algo, rho, rho_old, beta, alpha, Max;
	int iter, i;
	unsigned int timer_proc = 0;
	int send = Matrix.send;
	int recv = Matrix.recv;
	int threads = 512;
	int blocks = (sizeloc + threads - 1) / threads;
	int threads_send = 512;
	int blocks_send = (send + threads_send - 1) / threads_send;
	
	//Initialisation
	init(sizeloc, XSize, h_x_cg_cpu, h_b, h_p, h_q, h_r, h_z);
		
	//Matrix conversion from csr to hyp
	const int num_cols_per_row = compute_hyb_cols_per_row(csr);
	hyb_matrix<int,double> hyb = csr_to_hyb<int,double>(csr, num_cols_per_row);
	if(rank == rang)
		printf("\n\t\t -- ELL has : %d columns : \n\t\t -- NNZ : [ELL %4.1f%%] et [COO %4.1f%%]\n", num_cols_per_row,
				 100 * (double) hyb.ell.num_nonzeros / (double) csr.num_nonzeros, 
				 100 * (double) hyb.coo.num_nonzeros / (double) csr.num_nonzeros);
	
	const memory_location loc = DEVICE_MEMORY;
	hyb_matrix<int,double> sm_loc = (loc == DEVICE_MEMORY) ? copy_matrix_to_device(hyb) : hyb;

	double* decompressedX = new_host_array<double>(recv);   
	for(i=0; i<send; i++) Matrix.CompressedX[i] = 0.0;
	for(i=0; i<recv; i++) decompressedX[i] = 0.0;
	
	double * d_x = copy_array(h_x_cg_cpu, XSize, HOST_MEMORY, loc);
	double * d_b = copy_array(h_b, sizeloc, HOST_MEMORY, loc);
	double * d_d = copy_array(h_d, sizeloc, HOST_MEMORY, loc);
	double * d_r = copy_array(h_r, sizeloc, HOST_MEMORY, loc);       
	double * d_z = copy_array(h_z, sizeloc, HOST_MEMORY, loc);  
	double * d_q = copy_array(h_q, sizeloc, HOST_MEMORY, loc); 
	double * d_p = copy_array(h_p, XSize, HOST_MEMORY, loc); 
	double * d_compressedX = copy_array(Matrix.CompressedX, send, HOST_MEMORY, loc);
	int * d_SendColumns = copy_array(Matrix.SendColumns, send, HOST_MEMORY, loc);
	
	// Start of computation
	if(rank == rang){
		printf("\n\t\t\t\t ******   START OF GPU COMPUTATION   ******");
		printf("\n\t\t\t\t ******   START OF GPU COMPUTATION   ******\n");
	}
	CUDA_SAFE_CALL(cudaThreadSynchronize());
	MPI_Barrier(MPI_COMM_WORLD);
	CUT_SAFE_CALL(cutCreateTimer(&timer_proc));
  	CUT_SAFE_CALL(cutStartTimer(timer_proc));
	spmv_hyb_tex_device<int,double>(sm_loc, d_x, d_r);
	yminusx_kernel<<<blocks,threads>>>(sizeloc, d_b, d_r);
	cublasGetVector(sizeloc, sizeof(double), d_r, 1, h_r, 1);
	comp_norm(sizeloc, &normr, h_r);
	MPI_Allreduce(&normr, &norm_algo, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
	  	
	if(rank == rang)
		printf("\n\t\t -- GPU normr_begin : %20.20e", norm_algo);
	iter = 0;
		
	while(norm_algo>precision && iter<maxiter){
		s_kernel<<<blocks,threads>>>(sizeloc, d_q);
		diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_r, d_z);
		valeur = cublasDdot(sizeloc, d_r, 1, d_z, 1);
		MPI_Allreduce(&valeur, &rho, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
		
		if(iter == 0)
			cublasDcopy(sizeloc, d_z, 1, d_p, 1);
		else{
			beta = rho / rho_old;
			device_xeaxpy<<<blocks,threads>>>(sizeloc, beta, d_p, d_z);
		}	
		
		//CUDA_SAFE_CALL(cudaBindTexture(NULL, TextureX, d_p)); 
		Device_Compress_Vector<<<blocks_send,threads_send>>>(send, d_p, d_compressedX, d_SendColumns);
		//CUDA_SAFE_CALL(cudaUnbindTexture(TextureX)); 
		cublasGetVector(send, sizeof(double), d_compressedX, 1, Matrix.CompressedX, 1);
		Matrix.tag = 100;
		Synchronize(&Matrix, decompressedX);
		cublasSetVector(recv, sizeof(double), decompressedX, 1, &d_p[sizeloc], 1);
		spmv_hyb_tex_device<int,double>(sm_loc, d_p, d_q);

		valeur = cublasDdot(sizeloc, d_p, 1, d_q, 1);
		MPI_Allreduce(&valeur, &val, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

		alpha = rho / val;
		cublasDaxpy(sizeloc, alpha, d_p, 1, d_x, 1);
		cublasDaxpy(sizeloc, -alpha, d_q, 1, d_r, 1);
		rho_old = rho;
		iter ++;
		norm_algo = rho_old;
	}    	
	CUT_CHECK_ERROR("Kernel execution failed");
  	CUDA_SAFE_CALL(cudaThreadSynchronize());
	MPI_Barrier(MPI_COMM_WORLD);
	cutStopTimer(timer_proc);
	time_gpu = cutGetTimerValue(timer_proc);
	cutDeleteTimer(timer_proc);
	MPI_Allreduce(&time_gpu, &TIME_GPU, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	*cg_gpu_timer = TIME_GPU / nbprocs;
	if(rank == rang){
		printf("\n\t\t -- GPU processing time : %f (s)", *cg_gpu_timer/1000);
		printf("\n\t\t -- GPU iter_end : %d", iter);
		printf("\n\t\t -- GPU normb_end : %20.20e", norm_algo);
	}
	iteration = iter;
		
	s_kernel<<<blocks,threads>>>(sizeloc, d_r);
	//CUDA_SAFE_CALL(cudaBindTexture(NULL, TextureX, d_p)); 
	Device_Compress_Vector<<<blocks_send,threads_send>>>(send, d_x, d_compressedX, d_SendColumns);
	//CUDA_SAFE_CALL(cudaUnbindTexture(TextureX)); 
	cublasGetVector(send, sizeof(double), d_compressedX, 1, Matrix.CompressedX, 1);
	Matrix.tag = 101;
	Synchronize(&Matrix, decompressedX);
	cublasSetVector(recv, sizeof(double), decompressedX, 1, &d_x[sizeloc], 1);
	spmv_hyb_tex_device<int,double>(sm_loc, d_x, d_r);	
	yminusx_kernel<<<blocks,threads>>>(sizeloc, d_b, d_r);
	diag_solve_kernel<<<blocks,threads>>>(sizeloc, d_d, d_r, d_r);
	cublasGetVector(sizeloc, sizeof(double), d_r, 1, h_r, 1); 
	Max = 0.0; valeur = 0.0;
	for(i=0; i<sizeloc; i++)
    		if(Max < fabs(h_r[i]))
      			Max = fabs(h_r[i]); 
				 
  	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
	if(rank == rang)
		printf("\n\t\t -- GPU MAX of residu : %e\n\n", valeur);
	*h_x_cg_gpu = copy_array(d_x, sizeloc, loc, HOST_MEMORY);
		 
	//clean up device memory
	if(loc == DEVICE_MEMORY) 
		delete_device_matrix(sm_loc);

	delete_array(d_x, loc);
	delete_array(d_b, loc);
	delete_array(d_d, loc);
	delete_array(d_r, loc);
	delete_array(d_z, loc);
	delete_array(d_q, loc);
	delete_array(d_p, loc);
	free(decompressedX);
}
//----------------------------------------------------------------------------------------------------------------------------------------------------//

////////////////////////////////////////////////////////////////////////////////
//! Run a simple test for CUDA
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{
	int i, j, k, numrhs, reste, nonzeros, band, found; 
	char* matType;
	double gmres_cpu_timer, cg_cpu_timer, gmres_gpu_timer, cg_gpu_timer, Max, valeur, val;

//MPI INITIALIZATION	
	MPI_Init(&argc,&argv);
	MPI_Comm_size(MPI_COMM_WORLD, &nbprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

//CUDA INITIALIZATION
	cutilDeviceInit(argc, argv);	
	cudaSetDevice(rank % 2);
	cublasInit();

//	cudaDeviceProp deviceProp;
//	int device = rank % 2;
//	cudaGetDeviceProperties(&deviceProp, device);
//	printf("Device %d has concurrentKernels %d.\n", device, deviceProp.asyncEngineCount);


//PARAMETERS OF INITIALIZATION
	for(i=2; i<argc; i++){		
		if(strncmp(argv[i],"rank",4)==0)        rang = atoi(&(argv[i][4]));	//rang processus pour affichage
		if(strncmp(argv[i],"maxiter",7)==0)     maxiter = atoi(&(argv[i][7]));	//max iterations
		if(strncmp(argv[i],"prec",4)==0)        precision = atof(&(argv[i][4]));//precision
		if(strncmp(argv[i],"restart",7)==0)     restart = atoi(&(argv[i][7]));	//GMRES restarts
		if(strncmp(argv[i],"symmetric",9)==0)   symmetric = 1; 			//matrice symmetrique
		if(strncmp(argv[i],"method",6)==0)     	method = atoi(&(argv[i][6]));	//kind of matrix generation
		if(strncmp(argv[i],"nombre",6)==0)     	nombre = atoi(&(argv[i][6]));	//nombre de simulations
	}

//LOADING OF MATRIX
	readHB_info(argv[1], &MatrixA.NumRows, &MatrixA.NumColumns, &MatrixA.NumNonZeros, &matType, &numrhs);

	MatrixA.Pointers = (int*)malloc(sizeof(int)*(MatrixA.NumRows+1));
	MatrixA.Columns = (int*)malloc(sizeof(int)*MatrixA.NumNonZeros);
	MatrixA.NonZeros = (double*)malloc(sizeof(double)*MatrixA.NumNonZeros);
	readHB_mat_double(argv[1], MatrixA.Pointers, MatrixA.Columns, MatrixA.NonZeros);
	for(i=0; i<MatrixA.NumRows+1; i++) MatrixA.Pointers[i] --;
	for(i=0; i<MatrixA.NumNonZeros; i++) MatrixA.Columns[i] --;

//GENERATE THE UPPER PART IF SYMMETRIC
	Loaded_Matrix(&matrix, &MatrixA, &MatrixB, symmetric);

//MATRIX BANDWIDTH
	Bandwidth(&matrix);
	LargerRow(&matrix);
	Size_Sub_Copies(&matrix);

	if(rank == rang){
		printf("\n\t -------------------------  Loaded matrix : <<%s>>  -------------------------\n", argv[1]);
		printf("\n\t\t -- Number of rows: %d", matrix.NumRows);
		printf("\n\t\t -- Number of columns: %d", matrix.NumColumns);
		printf("\n\t\t -- Number of nonzero values: %d", matrix.NumNonZeros);
		printf("\n\t\t -- Bandwidth: %d", matrix.Bandwidth);
		printf("\n\t\t -- Cut row: %d", CutRow);
		printf("\n\t\t -- Left column: %d", LeftColumn);
		printf("\n\t\t -- Right column: %d", RightColumn);
		printf("\n\t\t -- Left portion: %d", LeftPortion);
		printf("\n\t\t -- Right portion: %d\n", RightPortion);
		printf("\n\t -----------------------------------------------------------------------------------------------------\n");
	}

//SIZE OF THE GLOBAL MATRIX
	sizeglob = matrix.NumRows;
	for(i=2; i<argc; i++){
		if(strncmp(argv[i],"matrixsize",10)==0){
			sizeglob = atoi(&(argv[i][10]));
			if(sizeglob < nbprocs){
				int idle = nbprocs - sizeglob;
				if(rank == rang) printf("\n\n\t\t -- The generated matrix size is smaller than the number of processes. We have %d idle processes\n\n", idle);
				MPI_Barrier(MPI_COMM_WORLD);
				MPI_Finalize();
				exit(0);
			}			
		}
	}

//SIZELOC & OFFSET
	Matrix.TabSizes = (int*)malloc(sizeof(int)*nbprocs);
	Matrix.Taboffsets = (int*)malloc(sizeof(int)*(nbprocs+1));
	sizeloc = sizeglob / nbprocs;
	reste = sizeglob - (sizeloc * nbprocs);
	offset = 0;
	for(i=0; i<nbprocs; i++){
		Matrix.TabSizes[i] = sizeloc;
		if(i < reste) Matrix.TabSizes[i] ++;
		Matrix.Taboffsets[i] = offset;
		offset += Matrix.TabSizes[i];
	}
	sizeloc = Matrix.TabSizes[rank];
	offset = Matrix.Taboffsets[rank];
	Matrix.Taboffsets[nbprocs] = sizeglob;

//NUMBER OF NONZEROS
	if(method == 0) estimation(&matrix, &Matrix);
	if(method == 1) estimation_1(&matrix, &Matrix);

//GENERATED MATRIX
	Matrix.NumRows = sizeloc;
	Matrix.NumColumns = sizeglob;
	Matrix.Pointers = (int*)malloc(sizeof(int)*(sizeloc+1));
	Matrix.Columns = (int*)malloc(sizeof(int)*(Matrix.NumNonZeros));
	Matrix.NonZeros = (double*)malloc(sizeof(double)*(Matrix.NumNonZeros));
	if(method == 0) Generation_Band_Sparse_Matrix(&matrix, &Matrix);
	if(method == 1) Generation_Sparse_Matrix_Of_Three_Bands(&matrix, &Matrix);

//LEFTECART & RIGHTECART
	Bandwidth(&Matrix);
	Computing_Borders(&Matrix);
	if(sizeloc != sizeglob){
		if(offset > LeftColumn) LeftEcart = offset - LeftColumn;
		else LeftEcart = 0;
		
		if(RightColumn >= (offset+sizeloc)) RightEcart = RightColumn - offset - sizeloc + 1;
		else RightEcart = 0;
	}
	else{
		LeftEcart = 0;
		RightEcart = 0;
	}

	if(rank == rang){
		printf("\n\t -------------------------  Local generated matrix  ---------------------------------------------------\n");
		printf("\n\t\t -- Number of rows: %d", Matrix.NumRows);
		printf("\n\t\t -- Number of columns: %d", Matrix.NumColumns);
		printf("\n\t\t -- Number of nonzero values: %d", Matrix.NumNonZeros);
		printf("\n\t\t -- Bandwidth: %d", Matrix.Bandwidth);
		printf("\n\t\t -- Left column: %d", LeftColumn);
		printf("\n\t\t -- Right column: %d", RightColumn);
		printf("\n\t\t -- Left ecart: %d", LeftEcart);
		printf("\n\t\t -- Right ecart: %d\n", RightEcart);
		printf("\n\t -----------------------------------------------------------------------------------------------------\n");
	}

//CHARACTERISTICS OF THE GLOBAL MATRIX
	MPI_Allreduce(&Matrix.NumNonZeros, &nonzeros, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
	MPI_Allreduce(&Matrix.Bandwidth, &band, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
	XSize = LeftEcart + RightEcart;
	int maxXSize; MPI_Allreduce(&XSize, &maxXSize, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);

	if(rank == rang){
		printf("\n\t -------------------------  Global generated matrix  ---------------------------------------------------\n");
		printf("\n\t\t -- Number of rows: %d", Matrix.NumColumns);
		printf("\n\t\t -- Number of columns: %d", Matrix.NumColumns);
		printf("\n\t\t -- Number of nonzero values: %d", nonzeros);
		printf("\n\t\t -- Bandwidth: %d", band);
		printf("\n\t\t -- Number of processes: %d", nbprocs);
		printf("\n\t\t -- Max XSize: %d\n", maxXSize);
		printf("\n\t -----------------------------------------------------------------------------------------------------\n");
	}

	//---------------------------------------------------------------------------------------------------------------------------------------------------------//
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//

//MATRIX CONVERSION FROM CSR TO CSC
	Conversion_CSR_CSC(&Matrix, &matrix);

//COMPUTE DATA DEPENDENCIES
	reste = 0; //number of nonzeros of the off matrix
	Matrix.sendcount = (int*)malloc(sizeof(int)*nbprocs);
	Matrix.sdispls   = (int*)malloc(sizeof(int)*nbprocs);
	Matrix.recvcount = (int*)malloc(sizeof(int)*nbprocs);
	Matrix.rdispls   = (int*)malloc(sizeof(int)*nbprocs);

	for(i=0; i<nbprocs; i++){
		Matrix.sendcount[i] = 0;
		Matrix.recvcount[i] = 0;
		Matrix.sdispls[i]   = 0;
		Matrix.rdispls[i]   = 0;
	}

//recvcounts
	for(i=0; i<nbprocs; i++){
		if(i != rank){
			for(j=Matrix.Taboffsets[i]; j<Matrix.Taboffsets[i]+Matrix.TabSizes[i]; j++){
				int nb = matrix.Pointers[j+1] - matrix.Pointers[j];
				reste += nb;
				if(nb != 0) Matrix.recvcount[i] ++;
			}
		}	
	}

//sendcounts
	MPI_Alltoall(Matrix.recvcount, 1, MPI_INT, Matrix.sendcount, 1, MPI_INT, MPI_COMM_WORLD);

//sdispls & rdispls
	int s, r;
	Matrix.ranksend = Matrix.rankrecv = s = r = 0;
	Matrix.send = Matrix.recv = 0;

	for(i=0; i<nbprocs; i++){
		if(Matrix.sendcount[i] != 0){
			Matrix.ranksend ++;
			Matrix.sdispls[i] = s;
			s += Matrix.sendcount[i];
			Matrix.send += Matrix.sendcount[i];
		}

		if(Matrix.recvcount[i] != 0){
			Matrix.rankrecv ++;
			Matrix.rdispls[i] = r;
			r += Matrix.recvcount[i];
			Matrix.recv += Matrix.recvcount[i];
		}
	}

//rankrecv & ranksend
	Matrix.RankSend = (int*)malloc(sizeof(int)*Matrix.ranksend);
	Matrix.RankRecv = (int*)malloc(sizeof(int)*Matrix.rankrecv);
	s = r = 0;

	for(i=0; i<nbprocs; i++){
		if(Matrix.sendcount[i] != 0){
			Matrix.RankSend[s] = i;
			s ++;
		}

		if(Matrix.recvcount[i] != 0){
			Matrix.RankRecv[r] = i;
			r ++;
		}
	}

	//---------------------------------------------------------------------------------------------------------------------------------------------------------//
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//

//MATRIX REORDERING
	MatrixA.Pointers = (int*)malloc(sizeof(int)*(matrix.NumRows+1));
	MatrixA.Columns = (int*)malloc(sizeof(int)*matrix.NumNonZeros);
	MatrixA.NonZeros = (double*)malloc(sizeof(double)*matrix.NumNonZeros);

	MatrixA.Pointers[0] = 0;
	s = 0;

//Local matrix
	for(i=Matrix.Taboffsets[rank]; i<Matrix.Taboffsets[rank]+Matrix.TabSizes[rank]; i++){
		for(j=matrix.Pointers[i]; j<matrix.Pointers[i+1]; j++){
			MatrixA.Columns[s] = matrix.Columns[j];
			MatrixA.NonZeros[s] = matrix.NonZeros[j];
			s ++; 
		}
		MatrixA.Pointers[i-Matrix.Taboffsets[rank]+1] = MatrixA.Pointers[i-Matrix.Taboffsets[rank]] + matrix.Pointers[i+1] - matrix.Pointers[i];
	}

//Off matrix
	r = Matrix.TabSizes[rank]; //next column index
	Matrix.RecvColumns = (int*)malloc(sizeof(int)*Matrix.recv);
	Matrix.SendColumns = (int*)malloc(sizeof(int)*Matrix.send);
	int count = 0;

	for(i=0; i<Matrix.rankrecv; i++){
		int RANK = Matrix.RankRecv[i]; 	
		for(j=Matrix.Taboffsets[RANK]; j<Matrix.Taboffsets[RANK]+Matrix.TabSizes[RANK]; j++){
			int nb = matrix.Pointers[j+1] - matrix.Pointers[j];
			if(nb != 0){
				for(k=matrix.Pointers[j]; k<matrix.Pointers[j+1]; k++){
					MatrixA.Columns[s] = matrix.Columns[k];
					MatrixA.NonZeros[s] = matrix.NonZeros[k];
					s ++; 
				}
				MatrixA.Pointers[r+1] = MatrixA.Pointers[r] + nb;
				r ++;
				Matrix.RecvColumns[count] = j;
				count ++;
			}
		}
	}

	free(matrix.Pointers);
	free(matrix.Columns);
	free(matrix.NonZeros);

//sendcolumns
	MPI_Alltoallv(Matrix.RecvColumns, Matrix.recvcount, Matrix.rdispls, MPI_INT, Matrix.SendColumns, Matrix.sendcount, Matrix.sdispls, MPI_INT, MPI_COMM_WORLD);
	for(i=0; i<Matrix.send; i++) Matrix.SendColumns[i] -= Matrix.Taboffsets[rank];

//Conversion from CSC to CSR
	XSize = MatrixA.NumRows = Matrix.TabSizes[rank] + Matrix.recv;
	MatrixA.NumColumns = matrix.NumColumns;
	MatrixA.NumNonZeros = matrix.NumNonZeros;
	Conversion_CSR_CSC(&MatrixA, &Matrix);
	Bandwidth(&Matrix);
	Computing_Borders(&Matrix);
	max_nonzeros(&Matrix);

	if(rank == rang){
		printf("\n\t -------------------------  Local generated matrix after reordering	----------------------------------\n");
		printf("\n\t\t -- Number of rows: %d", Matrix.NumRows);
		printf("\n\t\t -- Number of columns: %d", Matrix.NumColumns);
		printf("\n\t\t -- Number of nonzero values: %d", Matrix.NumNonZeros);
		printf("\n\t\t -- Bandwidth: %d", Matrix.Bandwidth);
		printf("\n\t\t -- Left column: %d", LeftColumn);
		printf("\n\t\t -- Right column: %d", RightColumn);
		printf("\n\t\t -- Number of data to send: %d", Matrix.send);
		printf("\n\t\t -- Number of data to receive: %d", Matrix.recv);
		printf("\n\t\t -- Max number of nonzeros in a matrix row: %d\n", Matrix.nmax);
		printf("\n\t -----------------------------------------------------------------------------------------------------\n");
	}

	//---------------------------------------------------------------------------------------------------------------------------------------------------------//
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//

	free(Matrix.TabSizes);
	free(Matrix.Taboffsets);
	free(Matrix.RecvColumns);
	Matrix.CompressedX = (double*)malloc(sizeof(double)*Matrix.send);
	Matrix.SendReq = (MPI_Request*)malloc(sizeof(MPI_Request)*nbprocs);
	Matrix.RecvReq = (MPI_Request*)malloc(sizeof(MPI_Request)*nbprocs);

//COMPRESSED SPARSE ROW
	csr_matrix<int,double> csr;
	csr.num_rows = Matrix.NumRows;
	csr.num_cols = Matrix.NumColumns;
	csr.num_nonzeros = Matrix.NumNonZeros;
	csr.Ap = Matrix.Pointers;
	csr.Aj = Matrix.Columns;
	csr.Ax = Matrix.NonZeros;

//PRECONDITIONNER
	double *h_d = (double*)malloc(sizeof(double)*sizeloc);
	for(i=0; i<sizeloc; i++){
		found = 0;
		for(j=Matrix.Pointers[i]; j<Matrix.Pointers[i+1]; j++){
			if(i == Matrix.Columns[j]){
				h_d[i] = 1.0 / Matrix.NonZeros[j];
				found = 1;
			}
		}
		if(!found) printf("\n\t\t --	PROBLEM	: %d	--\n",i);
	}
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//	

	if(rank == rang)
		printf("\n\t **********************     GENERALIZED MINIMAL RESIDUEL     **********************\n");

//ALLOCATE HOST MEMORY
	double *h_x_gmres_cpu = (double*) malloc(sizeof(double)*XSize);
	double *h_b = (double*) malloc(sizeof(double)*sizeloc);
	double *h_r = (double*) malloc(sizeof(double)*sizeloc);
	double *h_H = (double*) malloc(sizeof(double)*(restart+1)*(restart));
	double *h_v = (double*) malloc(sizeof(double)*(restart+1)*sizeloc);
	double *h_s = (double*) malloc(sizeof(double)*(restart+1));
	double *h_cs = (double*) malloc(sizeof(double)*(restart+1));
	double *h_sn = (double*) malloc(sizeof(double)*(restart+1));
	double *h_w = (double*) malloc(sizeof(double)*sizeloc);	
	double *h_tmp_V = (double*) malloc(sizeof(double)*XSize);
	double *h_tmp_v1 = (double*) malloc(sizeof(double)*(restart+1));
	double *h_tmp_v2 = (double*) malloc(sizeof(double)*(restart+1));
	double *h_x_gmres_gpu = (double*) malloc(sizeof(double)*sizeloc);

int test;
double test1=0.0, test2=0.0;
for(test=0; test<nombre; test++){
//GMRES on GPUs
	GMRES_On_GPU(csr, &h_x_gmres_gpu, &gmres_gpu_timer, h_x_gmres_cpu, h_b, h_r, h_d, h_H, h_v, h_s, h_cs, h_sn, h_w, h_tmp_V, h_tmp_v1, h_tmp_v2);
test1 += gmres_gpu_timer / 1000;
//GMRES on CPUs
	if(iteration != maxiter)
		GMRES_On_CPU(&gmres_cpu_timer, h_x_gmres_cpu, h_b, h_r, h_d, h_H, h_v, h_s, h_cs, h_sn, h_w, h_tmp_V, h_tmp_v1, h_tmp_v2);
test2 += gmres_cpu_timer / 1000;
	//Comparison between The CPU solution and the GPU solution
	Max = 0.0; valeur = 0.0;
	for(i=0; i<sizeloc; i++)
		if(Max<fabs(h_x_gmres_cpu[i]-h_x_gmres_gpu[i]))
			Max = fabs(h_x_gmres_cpu[i]-h_x_gmres_gpu[i]);
	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
	if(rank == rang){
		printf("\n\t\t -- MAXIMUM : %e", valeur);
                printf("\n\t\t -- SPEED UP : %f\n\n", (gmres_cpu_timer/gmres_gpu_timer));
	}
}	

	if(rank == rang){
		printf("\n\t\t **************** TIME GPU GMRES : %f (s)\n\n", (test1/nombre));
		printf("\n\t\t **************** TIME CPU GMRES : %f (s)\n\n", (test2/nombre));
		printf("\n\t\t **************** FINAL SPEED UP GMRES: %f\n\n", (test2/test1));
	}

	free(h_b);
	free(h_r);
	free(h_H);
	free(h_v);
	free(h_s);
	free(h_cs);
	free(h_w);
	free(h_tmp_V);
	free(h_tmp_v1);
	free(h_tmp_v2);

	//---------------------------------------------------------------------------------------------------------------------------------------------------------//
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//

	if(rank == rang)
		printf("\n\t **********************     CONJUGATE GRIADIENT     **********************\n");

//ALLOCATE HOST MEMORIES
	XSize = LeftEcart + sizeloc + RightEcart;
	double *h_x_cg_cpu = (double*) malloc(sizeof(double)*XSize);
	h_b = (double*) malloc(sizeof(double)*sizeloc);
	h_r = (double*) malloc(sizeof(double)*sizeloc);
	double *h_p = (double*) malloc(sizeof(double)*XSize);
	double *h_q =  (double*) malloc(sizeof(double)*sizeloc);
	double *h_z =  (double*) malloc(sizeof(double)*sizeloc);
	double *h_x_cg_gpu = (double*) malloc(sizeof(double)*sizeloc);

double test3=0.0, test4=0.0;
for(test=0; test<nombre; test++){
//CG on GPUs
	CG_On_GPU(csr, &h_x_cg_gpu, &cg_gpu_timer, h_x_cg_cpu, h_d, h_b, h_r, h_p, h_q, h_z);
test3 += cg_gpu_timer / 1000;
//CG on CPUs
	if(iteration != maxiter) 
		CG_On_CPU(&cg_cpu_timer, h_x_cg_cpu, h_d, h_b, h_r, h_p, h_q, h_z);
test4 += cg_cpu_timer / 1000;
	//Comparison between The CPU solution and the GPU solution
	Max = 0.0; valeur = 0.0;
	for(i=0; i<sizeloc; i++)
		if(Max<fabs(h_x_cg_cpu[i]-h_x_cg_gpu[i]))
			Max = fabs(h_x_cg_cpu[i]-h_x_cg_gpu[i]);
	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
	if(rank == rang){
		printf("\n\t\t -- MAXIMUM : %e", valeur);
                printf("\n\t\t -- SPEED UP : %f\n\n", (cg_cpu_timer/cg_gpu_timer));
	}
}
	if(rank == rang){
		printf("\n\t\t **************** TIME GPU CG : %f (s)\n\n", (test3/nombre));
		printf("\n\t\t **************** TIME CPU CG : %f (s)\n\n", (test4/nombre));
		printf("\n\t\t **************** FINAL SPEED UP CG : %f\n\n", (test4/test3));
	}
	
	free(h_b);
	free(h_r);
	free(h_q);
	free(h_p);
	free(h_z);

	//---------------------------------------------------------------------------------------------------------------------------------------------------------//
	//---------------------------------------------------------------------------------------------------------------------------------------------------------//

//Comparison between GMRES and CG on CPU
	Max = 0.0;
	for(i=0; i<sizeloc; i++)
		if(Max<fabs(h_x_cg_cpu[i]-h_x_gmres_cpu[i]))
			Max = fabs(h_x_cg_cpu[i]-h_x_gmres_cpu[i]);
	MPI_Allreduce(&Max, &valeur, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

//Comparison between GMRES and CG on GPU
	Max = 0.0;
	for(i=0; i<sizeloc; i++)
		if(Max<fabs(h_x_cg_gpu[i]-h_x_gmres_gpu[i]))
			Max = fabs(h_x_cg_gpu[i]-h_x_gmres_gpu[i]);
	MPI_Allreduce(&Max, &val, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

	if(rank == rang){
		printf("\n\n\t\t -- Diff GMRES & CG on CPU: %e", valeur);
		printf("\n\t\t -- Diff GMRES & CG on GPU: %e\n", val);
		printf("\n\t\t -- ** Speedup GMRES & CG on CPU: %f", (gmres_cpu_timer/cg_cpu_timer));
		printf("\n\t\t -- ** Speedup GMRES & CG on GPU: %f\n", (gmres_gpu_timer/cg_gpu_timer));
		printf("\n\t\t -- ++ Speedup CG && GMRES on CPU: %f", (cg_cpu_timer/gmres_cpu_timer));
		printf("\n\t\t -- ++ Speedup CG && GMRES on GPU: %f\n\n", (cg_gpu_timer/gmres_gpu_timer));
	}

	if(rank == rang){
		printf("\n\t\t **************** FINAL SPEED UP GPU: %f\n\n", (test1/test3));
		printf("\n\t\t **************** FINAL SPEED UP CPU: %f\n\n", (test2/test4));
	}

	free(h_d);
	free(h_x_gmres_cpu);
	free(h_x_gmres_gpu);
	free(h_x_cg_cpu);
	free(h_x_cg_gpu);

//END
	MPI_Finalize();
}

